import io
import base64
import pandas as pd
from mitosheet.sheet_functions.types.utils import get_mito_type


def get_column_summary_graph(send, event, wsc):
    """
    Sends back a string representation of a PNG graph of the column_header 
    in the df at sheet_index in the widget state container.

    If the series is a numeric series, will return a histogram. Otherwise, as
    long as there are less than 20 distinct items in the series, will return
    a bar chart of the value count. Otherwise, will return nothing.
    """
    sheet_index = event['sheet_index']
    column_header = event['column_header']
    
    series: pd.Series = wsc.dfs[sheet_index][column_header]

    try:
        mito_type = get_mito_type(series)

        if mito_type == 'number_series':
            ax = series.plot.hist()
        else:
            if series.nunique() > 20:
                # Send an empty response
                send({
                    'event': 'api_response',
                    'id': event['id'],
                    'data': ''
                })
                # Then return so we only respond with an empty string
                return
            ax = series.value_counts().plot.bar()
        ax.set_title(column_header + ' frequencies')

        # From here: https://stackoverflow.com/questions/38061267/matplotlib-graphic-image-to-base64
        pic_IObytes = io.BytesIO()
        ax.figure.savefig(pic_IObytes,  format='png', transparent=True, bbox_inches='tight')
        pic_IObytes.seek(0)
        pic_hash = base64.b64encode(pic_IObytes.read()).decode("utf-8").replace("\n", "")

        send({
            'event': 'api_response',
            'id': event['id'],
            'data': pic_hash
        })
    except Exception as e:
        # As not being able to make a graph is a non-critical error that doesn't
        # result from user interaction, we don't want to throw an error if something
        # weird happens, so we just return nothing in this case
        send({
            'event': 'api_response',
            'id': event['id'],
            'data': ''
        })


#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Mito.
# Distributed under the terms of the Modified BSD License.

"""
The Mito package, which contains functions for creating a Mito sheet. 

To generate a new sheet, simply run:

import mitosheet
mitosheet.sheet()

If running mitosheet.sheet() just prints text that looks like `MitoWidget(...`, then you need to 
install the JupyterLab extension manager by running:

jupyter labextension install @jupyter-widgets/jupyterlab-manager@2;

Run this command in the terminal where you installed Mito. It should take 5-10 minutes to complete.

Then, restart your JupyterLab instance, and refresh your browser. Mito should now render.

NOTE: if you have any issues with installation, please email book a demo time at https://hubs.ly/H0FL1920
"""

import os
import pandas as pd
from mitosheet.example import MitoWidget, sheet
from mitosheet.errors import CreationError, EditError
from mitosheet._version import __version__, version_info

# Export all the sheet functions
from mitosheet.sheet_functions import *
# And the functions for changing types
from mitosheet.sheet_functions.types import *

from .nbextension import _jupyter_nbextension_paths


if __name__ == 'mitosheet':
    from mitosheet.user.user import initialize_user, is_local_deployment
    # Make sure the user is created and up to date, whenever mitosheet is imported
    initialize_user()

    from mitosheet.utils import is_imported_correctly

    # We check local deployment first, as this hopefully short curcits and saves us time
    if is_local_deployment() and not is_imported_correctly():
        print("It looks like Mito is incorrectly installed. To finish installing Mito, follow the documentation here: https://docs.trymito.io/")


#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Mito.
# Distributed under the terms of the Modified BSD License.

"""
Contains helpful utility functions for getting and setting the user profile 
in user.json
"""

from copy import deepcopy
import os
import json

# Where all global .mito files are stored
MITO_FOLDER = os.path.expanduser("~/.mito")

# The path of the user.json file
USER_JSON_PATH = os.path.join(MITO_FOLDER, 'user.json')


def get_user_field(field):
    """
    Returns the value stored at field in the user.json file.
    """
    try:
        with open(USER_JSON_PATH) as f:
            return json.load(f)[field]
    except: 
        return None

def set_user_field(field, value):
    """
    Updates the value of a specific feild in user.json
    """
    with open(USER_JSON_PATH, 'r') as user_file_old:
        old_user_json = json.load(user_file_old)
        old_user_json[field] = value
        with open(USER_JSON_PATH, 'w+') as f:
            f.write(json.dumps(old_user_json))


#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Mito.
# Distributed under the terms of the Modified BSD License.

"""
Contains interfaces and helpers for dealing with the user profile.

Currently, the users profile is stored on disk where Mito is installed, 
but in the future we will support other profile storing locations (e.g.
on a server, so that users can log into a webapp!)
"""
import getpass
import os
import uuid
import json
from datetime import datetime

from mitosheet.mito_analytics import identify

from mitosheet._version import __version__
from mitosheet.user.user_utils import (
    get_user_field, set_user_field,
    MITO_FOLDER, USER_JSON_PATH
)
from mitosheet.mito_analytics import analytics



# This is the default user json object
USER_JSON_DEFAULT = {
    'user_json_version': 1,
    'static_user_id': '',
    'user_email': '',
    'mitosheet_current_version': __version__,
    'mitosheet_last_upgraded_date': datetime.today().strftime('%Y-%m-%d'),
    'mitosheet_last_five_usages': [datetime.today().strftime('%Y-%m-%d')]
}

def get_random_id():
    """
    Creates a new random ID for the user, which for any given user,
    should only happen once.
    """
    return str(uuid.uuid1())


def is_on_kuberentes_mito():
    """
    Returns True if the user is on Kuberentes Mito
    """
    user = getpass.getuser()
    return user == 'jovyan'


def is_local_deployment():
    """
    Helper function for figuring out if this a local deployment or a
    Mito server deployment
    """
    return not is_on_kuberentes_mito()  


def initialize_user():
    """
    Internal helper function that gets called every time mitosheet 
    is imported.

    It:
    1. Creates a ~/.mito folder, if it does not exist.
    2. Creates new, default user.json file if it does not exist, taking
       special care to do things properly if we're in a CI enviornment.
    3. Updates the user.json file with any new variables, as well as logging
       this usage, and any potential upgrades that may have occured
    """
    if not os.path.exists(MITO_FOLDER):
        os.mkdir(MITO_FOLDER)

    # We only create a user.json file if it does not exist
    if not os.path.exists(USER_JSON_PATH):
        # First, we write an empty default object
        with open(USER_JSON_PATH, 'w+') as f:
            f.write(json.dumps(USER_JSON_DEFAULT))

        # Then, we create a new static id and capture the email for the user. 
        # We take special care to put all the CI enviornments 
        # (e.g. Github actions) under one ID and email
        if 'CI' in os.environ and os.environ['CI'] is not None:
            static_user_id = 'github_action'
            user_email = 'github@action.com'
        else:
            static_user_id = get_random_id()
            # Then, we check if the user is on the server. If they are are on the server, then we
            # can get their email as their JUPYTERHUB_USER. If they are not on the server, then we set
            # their email as empty for now
            if is_on_kuberentes_mito():
                user_email = os.getenv('JUPYTERHUB_USER')
            else:
                user_email = ''

        set_user_field('static_user_id', static_user_id)
        set_user_field('user_email', user_email)


    # Then we just make sure that the user.json has all the fields it needs defined
    # and if they are not defined, it sets them to the default values
    for field, default_value in USER_JSON_DEFAULT.items():
        if get_user_field(field) is None:
            set_user_field(field, default_value)

    # Then, we check if Mito has been upgraded since it was last imported
    # and if it has been upgraded, we upgrade the version and the upgrade date
    if get_user_field('mitosheet_current_version') != __version__:
        set_user_field('mitosheet_current_version', __version__)
        set_user_field('mitosheet_last_upgraded_date', datetime.today().strftime('%Y-%m-%d'))

    # We also note this import as a Mito usage, making sure to only 
    # mark this as usage once per day
    last_five_usages = get_user_field('mitosheet_last_five_usages')
    if len(last_five_usages) == 0 or last_five_usages[-1] != datetime.today().strftime('%Y-%m-%d'):
        last_five_usages.append(datetime.today().strftime('%Y-%m-%d'))
    # Then, we take the 5 most recent (or as many as there are), and save them
    if len(last_five_usages) < 5:
        most_recent_five = last_five_usages
    else: 
        most_recent_five = last_five_usages[-5:]
    set_user_field('mitosheet_last_five_usages', most_recent_five)

    # Then, we identify the user, always
    identify()

def should_upgrade_mitosheet():
    """
    A helper function that calculates if a user should upgrade,
    which in practice does this with the following heuristic:
    1. If the user has not upgraded in two weeks, then we always reccomend that the user
       upgrades.
    2. If the user has used the tool 4 times since they last upgraded, then we also reccomend
       that they upgrade

    NOTE: this should always return false if it is not a local installation, for obvious
    reasons.

    The motivation here is just: we want them to upgrade frequently, but we also don't
    want to just bombard them with upgrade messages. This is a nice middle ground.
    """
    if not is_local_deployment():
        return False

    mitosheet_last_upgraded_date = datetime.strptime(get_user_field('mitosheet_last_upgraded_date'), '%Y-%m-%d')
    mitosheet_last_five_usages = [datetime.strptime(usage, '%Y-%m-%d') for usage in get_user_field('mitosheet_last_five_usages')]

    current_time = datetime.now()
    # Condition (1)
    if (current_time - mitosheet_last_upgraded_date).days > 14:
        return True
    # Condition (2)
    elif len(mitosheet_last_five_usages) >= 4:
        # As this list is chronological, we just need to check 4 back
        if mitosheet_last_five_usages[-4] > mitosheet_last_upgraded_date:
            return True

    return False


# Copyright (c) Mito.
# Distributed under the terms of the Modified BSD License.

# AUTOGENERATED | DO NOT EDIT BELOW
version_info = (0, 1, 158)
# AUTOGENERATED | DO NOT EDIT ABOVE
__version__ = ".".join(map(str, version_info))


#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Mito.
# Distributed under the terms of the Modified BSD License.

"""
File that contains utilities for deploying a new version of Mito to 
app.trymito.io. See main for running instructions
"""
import sys
import subprocess
from string import Template
import json
from time import sleep
import os
from string import Template, ascii_lowercase
import random
from typing import List

def safe_run_command(command: List[str]):
    """
    Helper function for running a command in the terminal. Throws
    an exception of the command returns an exception. Returns a tuple
    of (stdout, stderr).
    """

    results = subprocess.run(
        command,
        stdout=subprocess.PIPE, 
        stderr=subprocess.PIPE,
        universal_newlines=True
    )

    if results.returncode != 0:
        raise Exception(f'Exception running {" ".join(command)}:', results.stdout, results.stderr)

    return results.stdout, results.stderr


def build_dockerfile_from_template(mito_version) -> str:
    """
    Helper function that builds a Mito dockerfile from the template
    in deployment/templates/Dockerfile.

    Notably, includes all the files in tutorial in the dockerfile itself
    to be put on the users account. 
    """

    # Make sure no folders in the tutorial (for now)
    root_dir = 'deployment/tutorial'
    tutorial_files = set()

    for dir_, _, files in os.walk(root_dir):
        for file_name in files:
            rel_dir = os.path.relpath(dir_, root_dir)
            rel_file = os.path.join(rel_dir, file_name)
            tutorial_files.add(rel_file)

    if '.DS_Store' in tutorial_files:
        tutorial_files.remove('.DS_Store') 
    
    tutorial_copies = '\n'.join([f'COPY {f} /tutorial/{f}' for f in tutorial_files])

    # Read in template, replace with version and the tutorial
    with open('deployment/templates/docker_template.txt', 'r') as f:
        return Template(f.read()).substitute({
            'version': mito_version,
            'tutorial_copies': tutorial_copies
        })

def auth_with_aws():
    """
    Authenticates the docker CLI with AWS, so we can push to AWS ECR
    """
    print("Authenticating with AWS...")
    aws_ecr = subprocess.Popen(["aws", "ecr", "get-login-password", "--region", "us-east-1"], stdout=subprocess.PIPE)
    output = subprocess.check_output(["docker", "login", "--username", "AWS", "--password-stdin", "308734690614.dkr.ecr.us-east-1.amazonaws.com"], stdin=aws_ecr.stdout) # pipe them together
    aws_ecr.wait()


def build_and_upload_docker_image(mito_version):
    """
    Builds the docker image defined by the template docker file in templates, 
    as well as the files in the tutorial, and deploys it to AWS ECR.
    """
    dockerfile_string = build_dockerfile_from_template(mito_version)
    with open('deployment/tutorial/Dockerfile', 'w+') as f:
        f.write(dockerfile_string)
    
    # Go into the folder, and build the docker image
    print('Building a new docker image... this might take a bit.')
    os.chdir('deployment/tutorial')
    try:
        stdout, stderr = safe_run_command(["docker", "build", "."])
    except:
        # If there is an error during build, remove the dockerfile from the tutorial
        os.remove('Dockerfile')
        # and raise the error
        raise
    os.chdir('../..')

    # Remove the docker file
    os.remove('deployment/tutorial/Dockerfile')

    # Get the image id to actually deploy it to AWS ECR. 
    # stdout looks like:
    # Step 5/5 : COPY test.txt /tutorial/test.txt
    # ---> d6e2cb5b2899
    # Successfully built d6e2cb5b2899 
    image_id = [line for line in stdout.split('\n') if line.strip()][-1].strip().split(' ')[-1].strip()

    # Authenticate with ECR, so we can push the new image
    auth_with_aws()
    
    # NOTE: we add a random tag onto the end of the version tag, so that we get a new version if we redeploy it.
    # This makes the configs differ, and so ensures that JHub restarts
    tag_end = ''.join(random.choice(ascii_lowercase) for i in range(10))
    tag = f'{mito_version}.{tag_end}'

    # Tag the new image locally
    safe_run_command(
        ["docker", "tag", image_id, f'308734690614.dkr.ecr.us-east-1.amazonaws.com/mito-images:{tag}']
    )

    # And push the image
    print("Uploading the image to AWS... this might take a bit.")
    safe_run_command(
        ["docker", "push", f'308734690614.dkr.ecr.us-east-1.amazonaws.com/mito-images:{tag}']
    )
    return tag

def get_package_version() -> str:
    with open('package.json') as f:
        return json.loads(f.read())['version']

def check_files_on_version(version):
    """
    A helper function that checks that the version has been updated properly
    in all the correct places:
    - package.json
    - mitosheet/_version.py

    If the verison is not correctly updated, returns the offending file
    """
    files_to_check = [
        'package.json',
    ]

    for file_name in files_to_check:
        with open(file_name) as f:
            if version not in f.read():
                print(f'File {file_name} not updated to version {version}')
                return False
    
    # The mitosheet/_version.py file stores the version in a tuple, and so we read from that
    version_info_tuple = ', '.join(version.split('.'))
    with open('mitosheet/_version.py') as f:
        if version_info_tuple not in f.read():
            print(f'File mitosheet/_version.py not updated to version {version}')
            return False
    
    return True

def deploy_current_mito_version_to_pypi():
    """
    Deploys the current local version of Mito to PyPi.
    """
    deploy_results = subprocess.run(
        ["python3", "setup.py", "sdist", "bdist_wheel", "upload"],
        stdout=subprocess.PIPE, 
        stderr=subprocess.PIPE,
        universal_newlines=True
    )
    if deploy_results.returncode != 0:
        raise Exception("Failed to deploy to PyPi with output:", deploy_results.stdout, deploy_results.stderr)


def deploy_version_on_staging(tag):
    """
    Deploys the new version of staging!
    """
    print("Deploying on staging")
    
    # First, switch to the correct cluster!
    safe_run_command(
        ["aws", "eks", "--region", "us-east-1", "update-kubeconfig", "--name", "staging"]
    )

    # First, read in the secret token for the config
    with open('deployment/secrets.json', 'r') as f:
        secret_token = json.loads(f.read())['staging']['secretToken']

    with open('deployment/templates/template-staging-config.yaml', 'r') as f:
        staging_config_string = Template(f.read()).substitute({
            'secretToken': secret_token,
            'tag': tag
        })

    with open('staging-config.yaml', 'w+') as f:
        print("Writing...")
        f.write(staging_config_string)

    safe_run_command(
        ["helm", "upgrade", "--cleanup-on-fail", "--install", "jhub", "jupyterhub/jupyterhub", "--version", "0.10.2", "--values", "staging-config.yaml"]
    )

    # Note: we leave the config files there so that you can see what was deployed!

    print("Run \'aws eks --region us-east-1 update-kubeconfig --name staging; kubectl get pods\' to see the progress of this update.")


def deploy_version_on_app(tag):
    """
    Deploys the new version of the app on app!
    """
    print("Deploying on app")

    # First, switch to the right cluster
    safe_run_command(
        ["aws", "eks", "--region", "us-east-1", "update-kubeconfig", "--name", "app"]
    )

    # First, read in the secret token for the config
    with open('deployment/secrets.json', 'r') as f:
        secrets = json.loads(f.read())['app']
        secret_token = secrets['secretToken']
        client_id = secrets['google']['clientId']
        client_secret = secrets['google']['clientSecret']

    with open('deployment/templates/template-config.yaml', 'r') as f:
        config_string = Template(f.read()).substitute({
            'secretToken': secret_token,
            'tag': tag,
            'clientId': client_id,
            'clientSecret': client_secret
        })

    with open('config.yaml', 'w+') as f:
        f.write(config_string)

    safe_run_command(
        ["helm", "upgrade", "--cleanup-on-fail", "--install", "jhub", "jupyterhub/jupyterhub", "--version", "0.10.2", "--values", "config.yaml"]
    )

    # Note: we leave the config files there so that you can see what was deployed!

    print("Run \'kubectl get pods\' to see the progress of this update.")


def main():
    """
    This deploys the current Mito version on either staging, app, or all. Simply
    run this script with `python3 deployment/deploy.py [staging | app | all]`.

    To change the tutorial that new users get, simply change the files in the tutorial folder.
    """

    # We either deploy app, staging, or all (both app and staging)
    deploy_location = sys.argv[1]
    if deploy_location not in ['app', 'staging', 'all']:
        raise Exception(f'Invalid deploy location: {deploy_location}. Please choose from app, staging, or all.')

    # Make sure all the files have been updated to the right version
    mito_version = get_package_version()
    if not check_files_on_version(mito_version):
        raise Exception(f'At least one file has not been updated to the proper version. Please update!')

    # Then, we actually deploy Mito to PyPi, if it has not been deployed yet
    try:
        print(f'Trying to deploy Mito {mito_version} on PyPi... might take a moment')
        deploy_current_mito_version_to_pypi()
    except Exception as e:
        print(e)
        print("Failed to deploy to PyPi. Assuming this is because you deployed already, so continuing... but we may be doomed.")

    # We wait, just in case ECR needs to update
    print("Sleeping for 3 minutes, to allow things to update on PyPi...")
    sleep(60 * 3)

    # Then, we build a new Mito image on Docker to AWS ECR
    docker_tag = build_and_upload_docker_image(mito_version)

    # We wait, just in case ECR needs to update
    print("Sleeping for 20 seconds, to allow things to update on AWS's end...")
    sleep(20)

    # Finially, we case and deploy
    if deploy_location == 'staging':
        deploy_version_on_staging(docker_tag)
    if deploy_location == 'app':
        deploy_version_on_app(docker_tag)
    if deploy_location == 'all':
        deploy_version_on_staging(docker_tag)
        deploy_version_on_app(docker_tag)

    # Always switch back to the correct cluster!
    safe_run_command(
        ["aws", "eks", "--region", "us-east-1", "update-kubeconfig", "--name", "app"]
    )

if __name__ == '__main__':
    main()


#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Mito.
# Distributed under the terms of the Modified BSD License.

"""
A signup update, sets the user's email as the user_email in user.json
"""

from mitosheet.mito_analytics import identify
from mitosheet.user.user_utils import set_user_field


SIGNUP_EVENT = 'signup_update'
SIGNUP_PARAMS = ['user_email']


def execute_signup_update(wsc, user_email):
    """
    The function responsible for signing in the user.
    """
    # Set the user_email in user.json
    set_user_field('user_email', user_email)

    # Identify the user with their new email
    identify()

SIGNUP_UPDATE = {
    'event_type': SIGNUP_EVENT,
    'params': SIGNUP_PARAMS,
    'execute': execute_signup_update
}


#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Mito.
# Distributed under the terms of the Modified BSD License.

"""
After reading in the dataframe names from the frontend, 
sets their name in the widget state container
"""

from copy import deepcopy


DF_NAMES_UPDATE_EVENT = 'df_names_update'
DF_NAMES_UPDATE_PARAMS = ['df_names']

def execute_df_names_update(
        wsc,
        df_names
    ):
    """
    Changes the df_names in the current step - from information
    from the frontend
    """
    wsc.curr_step['df_names'] = deepcopy(df_names)


DF_NAMES_UPDATE = {
    'event_type': DF_NAMES_UPDATE_EVENT,
    'params': DF_NAMES_UPDATE_PARAMS,
    'execute': execute_df_names_update
}


#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Mito.
# Distributed under the terms of the Modified BSD License.

"""
Contains utilities for bumping the version of a Mito project, across
all the files where it needs to be bumped. 
"""

import json
from typing import Tuple

def get_package_version() -> Tuple[int, int, int]:
    with open('package.json') as f:
        str_version =  json.loads(f.read())['version']

    return tuple(map(int, str_version.split('.')))

def get_next_version() -> Tuple[int, int, int]:
    (x, y, z) = get_package_version()
    return (x, y, z + 1)

def bump_version(new_version: Tuple[int, int, int]=None):
    """
    Bumps the version of the Mito project to the next minor logical version.

    If a new_version is given, then will bump to that version specificially. new_version
    should be a tuple of the form (x, y, z).
    """

    if new_version is None:
        new_version = get_next_version()

    # Open mitosheet/_version.py, and replace the version
    with open('mitosheet/_version.py', 'r+') as py_verson:
        py_version_contents = py_verson.read()
        """
        The file has the format:

        # AUTOGENERATED | DO NOT EDIT BELOW
        version_info = (x, y, z)
        # AUTOGENERATED | DO NOT EDIT ABOVE
        """
        AUTOGENERATED_START = '# AUTOGENERATED | DO NOT EDIT BELOW\n'
        AUTOGENERATED_END = '# AUTOGENERATED | DO NOT EDIT ABOVE\n'

        start = py_version_contents.find(AUTOGENERATED_START)
        end = py_version_contents.find(AUTOGENERATED_END) + len(AUTOGENERATED_END)

        next_version_info = f'version_info = {new_version}\n'
        new_contents = py_version_contents[:start] + AUTOGENERATED_START + next_version_info + AUTOGENERATED_END + py_version_contents[end:]

        # Go back to start of file, to write
        py_verson.seek(0)
        py_verson.write(new_contents)    
    
    # Then, update it in package.json
    with open('package.json', 'r+') as f:
        package_obj = json.loads(f.read())
        package_obj['version'] = '.'.join(map(str, new_version))

        # Rewind to start of file, to write
        f.seek(0)
        f.write(json.dumps(package_obj, indent=2))

    print('Bump Mito version to ', new_version)

if __name__ == '__main__':
    bump_version()


#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Jupyter Development Team.
# Distributed under the terms of the Modified BSD License.

from __future__ import print_function
from glob import glob
from os.path import join as pjoin


from setupbase import (
    create_cmdclass, install_npm, ensure_targets,
    find_packages, combine_commands, ensure_python,
    get_version, HERE
)

from setuptools import setup


# The name of the project
name = 'mitosheet'

# Ensure a valid python version
ensure_python('>=3.4')

# Get our version
version = get_version(pjoin(name, '_version.py'))

nb_path = pjoin(HERE, name, 'nbextension', 'static')
lab_path = pjoin(HERE, name, 'labextension')

# Representative files that should exist after a successful build
jstargets = [
    pjoin(nb_path, 'index.js'),
    pjoin(HERE, 'lib', 'plugin.js'),
]

package_data_spec = {
    name: [
        'nbextension/static/*.*js*',
        'labextension/*.tgz'
    ]
}

data_files_spec = [
    ('share/jupyter/nbextensions/mitosheet',
        nb_path, '*.js*'),
    ('share/jupyter/lab/extensions', lab_path, '*.tgz'),
    ('etc/jupyter/nbconfig/notebook.d' , HERE, 'mitosheet.json')
]


cmdclass = create_cmdclass('jsdeps', package_data_spec=package_data_spec,
    data_files_spec=data_files_spec)
cmdclass['jsdeps'] = combine_commands(
    install_npm(HERE, build_cmd='build:all'),
    ensure_targets(jstargets),
)


setup_args = dict(
    name            = name,
    description     = 'The Mito Spreadsheet',
    version         = version,
    scripts         = glob(pjoin('scripts', '*')),
    cmdclass        = cmdclass,
    packages        = find_packages(),
    author          = 'Mito',
    author_email    = 'naterush1997@gmail.com',
    url             = 'https://github.com/mito/mito',
    license         = 'BSD',
    platforms       = "Linux, Mac OS X, Windows",
    keywords        = ['Jupyter', 'Widgets', 'IPython'],
    classifiers     = [
        'Intended Audience :: Developers',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: BSD License',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Framework :: Jupyter',
    ],
    include_package_data = True,
    install_requires = [
        # We require jupyterlab 2.0
        'jupyterlab>=2.0,<3.0',
        'ipywidgets>=7.0.0',
        'pandas>=1.1.0',
        'matplotlib>=3.3',
        # We don't need to lock an analytics-python version, as this library
        # is stabilized and mature
        'analytics-python'
    ],
    extras_require = {
        'test': [
            'pytest>=4.6',
            'pytest-cov',
            'nbval',
        ],
        'examples': [
            # Any requirements for the examples to run
        ],
        'docs': [
            'sphinx>=1.5',
            'recommonmark',
            'sphinx_rtd_theme',
            'nbsphinx>=0.2.13,<0.4.0',
            'jupyter_sphinx',
            'nbsphinx-link',
            'pytest_check_links',
            'pypandoc',
        ],
    },
    entry_points = {
    },
    long_description="""
        To learn more about Mito, checkout out our documentation: https://docs.trymito.io/getting-started/installing-mito\n\n
        Before installing Mito \n\n
        1. Check that you have Python 3.6 or above. To check your version of Python, open a new terminal, and type python3 --version. If you need to install or update Python, restart your terminal after doing so.\n\n
        2. Check that you have Node installed.To check this, open a new terminal, and type node -v.  It should print a version number. If you need to install Node, restart your terminal after doing so.\n\n
        3. Mito works in Jupyter Lab 2.0 only. We do not yet support Google Collab, VSCode, or Jupyter Lab 3.0.\n\n
        4. Checkout our terms of service and privacy policy. By installing Mito, you're agreeing to both of them. Please contact us at aarondr77 (@) gmail.com with any questions.\n\n
        Installation Instructions \n\n
        For more detailed installation instructions, see our documentation: https://docs.trymito.io/getting-started/installing-mito\n\n
        1. pip install mitosheet\n\n
        2. jupyter labextension install @jupyter-widgets/jupyterlab-manager@2\n\n
        3. jupyter lab
    """,
    long_description_content_type='text/markdown'
)

if __name__ == '__main__':
    setup(**setup_args)



#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Mito.
# Distributed under the terms of the Modified BSD License.

"""
TODO: Add module docstring
"""
import json
from mitosheet.api import handle_api_event
from mitosheet.mito_analytics import get_dfs_metadata, log, log_event_processed, log_recent_error
from mitosheet.steps import STEPS
import pandas as pd
from typing import List

from ipywidgets import DOMWidget
import traitlets as t

from mitosheet._frontend import module_name, module_version
from mitosheet.errors import EditError, get_recent_traceback_as_list
from mitosheet.save_utils import (
    read_and_upgrade_analysis, 
    write_analysis,
    saved_analysis_names_json
)

from mitosheet.widget_state_container import WidgetStateContainer
from mitosheet.profiling import timeit

from mitosheet.user.user import is_local_deployment, should_upgrade_mitosheet

class MitoWidget(DOMWidget):
    """
        The MitoWidget holds all of the backend state for the Mito extension, and syncs
        the state with the frontend widget. 
    """
    _model_name = t.Unicode('ExampleModel').tag(sync=True)
    _model_module = t.Unicode(module_name).tag(sync=True)
    _model_module_version = t.Unicode(module_version).tag(sync=True)
    _view_name = t.Unicode('ExampleView').tag(sync=True)
    _view_module = t.Unicode(module_name).tag(sync=True)
    _view_module_version = t.Unicode(module_version).tag(sync=True)

    is_local_deployment = t.Bool(True).tag(sync=True)
    analysis_name = t.Unicode('').tag(sync=True)
    curr_step_idx = t.Int(0).tag(sync=True)
    sheet_json = t.Unicode('').tag(sync=True)
    code_json = t.Unicode('').tag(sync=True)
    df_names_json = t.Unicode('').tag(sync=True)
    df_shape_json = t.Unicode('').tag(sync=True)
    saved_analysis_names_json = t.Unicode('').tag(sync=True)
    column_spreadsheet_code_json = t.Unicode('').tag(sync=True)
    column_filters_json = t.Unicode('').tag(sync=True)
    column_type_json = t.Unicode('').tag(sync=True)
    tutorial_mode = t.Bool(False).tag(sync=True)
    has_rendered = t.Bool(True).tag(sync=True)
    user_email = t.Unicode('').tag(sync=True)
    step_data_list_json = t.Unicode('').tag(sync=True)
    should_upgrade_mitosheet = t.Bool(False).tag(sync=True)

    def __init__(self, *args, **kwargs):
        """
        Takes a list of dataframes, passed through *args. 

        NOTE: assumes that the passed arguments are all dataframes, 
        and also have all valid headers. These conditions are checked in the
        mitosheet.sheet function. 
        """
        # Call the DOMWidget constructor to set up the widget properly
        super(MitoWidget, self).__init__()

        # Set if this is a local deployment
        self.is_local_deployment = is_local_deployment()

        # Mark if it is time for the user to update
        self.should_upgrade_mitosheet = should_upgrade_mitosheet()
            
        # Set up the state container to hold private widget state
        self.widget_state_container = WidgetStateContainer(args)

        # Set the widget's tutorial mode
        self.tutorial_mode = kwargs['tutorial_mode']
        
        # When the widget is first created, it has not been rendered on the frontend yet,
        # but after it is rendered once, this is set to False. This helps us detect 
        # when we are rendering for the first time vs. refreshing the sheet
        self.has_rendered = False

        # Set up starting shared state variables
        self.update_shared_state_variables()

        # Set up message handler
        self.on_msg(self.receive_message)

    def update_shared_state_variables(self):
        """
        Helper function for updating all the variables that are shared
        between the backend and the frontend through trailets.
        """
        self.sheet_json = self.widget_state_container.sheet_json
        self.curr_step_idx = self.widget_state_container.curr_step_idx
        self.column_spreadsheet_code_json = self.widget_state_container.column_spreadsheet_code_json
        self.code_json = self.widget_state_container.code_json
        self.df_names_json = self.widget_state_container.df_names_json
        self.df_shape_json = self.widget_state_container.df_shape_json
        self.analysis_name = self.widget_state_container.analysis_name
        self.saved_analysis_names_json = saved_analysis_names_json()
        self.column_filters_json = self.widget_state_container.column_filters_json
        self.column_type_json = self.widget_state_container.column_type_json
        self.user_email = self.widget_state_container.user_email
        self.step_data_list_json = self.widget_state_container.step_data_list_json


    def send(self, msg):
        """
        We overload the DOMWidget's send function, so that 
        we log all outgoing messages
        """
        # Send the message though the DOMWidget's send function
        super().send(msg)    
        

    def handle_edit_event(self, event):
        """
        Handles an edit_event. Per the spec, an edit_event
        updates both the sheet and the codeblock, and as such
        the sheet is re-evaluated and the code for the codeblock
        is re-transpiled.

        Useful for any event that changes the state of both the sheet
        and the codeblock!
        """
        # First, we send this new edit to the evaluator
        self.widget_state_container.handle_edit_event(event)

        # We update the state variables 
        self.update_shared_state_variables()

        # Also, write the analysis to a file!
        write_analysis(self.widget_state_container)

        # Tell the front-end to render the new sheet and new code
        self.send({"event": "update_sheet"})
        self.send({"event": "update_code"})


    def handle_update_event(self, event):
        """
        This event is not the user editing the sheet, but rather information
        that has been collected from the frontend (after render) that is being
        passed back.

        For example:
        - Names of the dataframes
        - Name of an existing analysis
        """
        # If this is just a message to notify the backend that we have rendered, set and return
        if event['type'] == 'has_rendered_update':
            self.has_rendered = True
            return

        self.widget_state_container.handle_update_event(event)

        # Update all state variables
        self.update_shared_state_variables()

        # Also, write the analysis to a file!
        write_analysis(self.widget_state_container)

        # Tell the front-end to render the new sheet and new code
        self.send({"event": "update_sheet"})
        self.send({"event": "update_code"})

    def receive_message(self, widget, content, buffers=None):
        """
        Handles all incoming messages from the JS widget. There are two main
        types of events:

        1. edit_event: any event that updates the state of the sheet and the
        code block at once. Leads to reevaluation, and a re-transpile.

        2. update_event: any event that isn't caused by an edit, but instead
        other types of new data coming from the frontend (e.g. the df names 
        or some existing steps).

        3. A log_event is just an event that should get logged on the backend.
        """
        event = content

        try:
            if event['event'] == 'edit_event':
                self.handle_edit_event(event)
            elif event['event'] == 'update_event':
                self.handle_update_event(event)
            elif event['event'] == 'api_call':
                handle_api_event(self.send, event, self.widget_state_container)
                return 
            
            # NOTE: we don't need to case on log_event above because it always gets
            # passed to this function, and thus is logged. However, we do not log
            # api calls, as they are just noise.
            log_event_processed(event, self.widget_state_container)
        except EditError as e:
            print(get_recent_traceback_as_list())
            print(e)
            
            # Log processing this event failed
            log_event_processed(event, self.widget_state_container, failed=True, edit_error=e)

            # Report it to the user, and then return
            self.send({
                'event': 'edit_error',
                'type': e.type_,
                'header': e.header,
                'to_fix': e.to_fix
            })
        except:
            print(get_recent_traceback_as_list())
            # We log that processing failed, but have no edit error
            log_event_processed(event, self.widget_state_container, failed=True)
            # Report it to the user, and then return
            self.send({
                'event': 'edit_error',
                'type': 'execution_error',
                'header': 'Execution Error',
                'to_fix': 'Sorry, there was an error during executing this code.'
            })

def sheet(
        *args: List[pd.DataFrame], 
        tutorial_mode=False,
        saved_analysis_name=None
    ) -> MitoWidget:
    """
    Renders a Mito sheet. If no arguments are passed, renders an empty sheet. Otherwise, renders
    any dataframes that are passed. Errors if any given arguments are not dataframes.

    If running this function just prints text that looks like `MitoWidget(...`, then you need to 
    install the JupyterLab extension manager by running:

    jupyter labextension install @jupyter-widgets/jupyterlab-manager@2;

    Run this command in the terminal where you installed Mito. It should take 5-10 minutes to complete.

    Then, restart your JupyterLab instance, and refresh your browser. Mito should now render.

    NOTE: if you have any issues with installation, please book a demo at https://hubs.ly/H0FL1920
    """
    args = list(args)

    # First, we validate all of the parameters passed to the sheet function, and throw an
    # error if any of them are invalid
    passed_df = False
    for df in args:
        # If it's a string, we assume the user passed a filename, and so 
        # report a special error to them telling them to not pass file names
        if isinstance(df, str):
            log(
                'mitosheet_sheet_call_failed', 
                {'error': f'Invalid argument passed to sheet: {df}. Instead of passing a filename, use pandas to read in the file as a dataframe and then pass the resulting dataframe to this function.'}
            )
            raise ValueError(f'Invalid argument passed to sheet: {df}. Instead of passing a filename, use pandas to read in the file as a dataframe and then pass the resulting dataframe to this function.')
        
        if not isinstance(df, pd.DataFrame):
            # Log the error
            log(
                'mitosheet_sheet_call_failed', 
                {'error': f'Invalid argument passed to sheet: {df}. Please pass all dataframes.'}
            )

            raise ValueError(f'Invalid argument passed to sheet: {df}. Please pass all dataframes.')
        else:
            passed_df = True

    # If they passed a dataframe, assume its personal data and log this,
    # unless its the data that we provide in the docs
    if passed_df and args[0].columns.tolist() != ['Zip', 'City', 'State', 'Ticket_Office']:
        log('used_personal_data')    

    if saved_analysis_name is not None and read_and_upgrade_analysis(saved_analysis_name) is None:
        log(
            'mitosheet_sheet_call_failed', 
            {'error': f'There is no saved analysis with the name {saved_analysis_name}.'}
        )
        raise ValueError(f'There is no saved analysis with the name {saved_analysis_name}.')

    try:
        widget = MitoWidget(*args, tutorial_mode=tutorial_mode) 
    except:
        # We log the error
        log_recent_error('mitosheet_sheet_call_failed')
        # And then bubble it to the user
        raise

    # Then, we log that the call was successful, along with all of it's params
    log(
        'mitosheet_sheet_call',
        dict(
            **{
                'param_tutorial_mode': tutorial_mode,
                'param_saved_analysis_name': saved_analysis_name
            },
            **{
                # NOTE: analysis name is the UUID that mito saves the analysis under
                'wsc_analysis_name': widget.widget_state_container.analysis_name,
            },
            **get_dfs_metadata(args)
        )
    )
    
    # If an analysis is passed, we pass this to the widget
    if saved_analysis_name is not None:
        widget.receive_message(widget, {
            'event': 'update_event',
            'type': 'replay_analysis_update',
            'analysis_name': saved_analysis_name,
            'import_summaries': None,
            'clear_existing_analysis': False
        })

    return widget



#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Mito.
# Distributed under the terms of the Modified BSD License.

"""
Our general approach to logging can be understood as:
0. Users are identified by a single random ID that we generate on the client-side, 
   here. We store this in the ~/.mito/user.json file, and we use it as the permanent
   ID for all users.
1. All logging is on the backend. This avoids us worrying about blockers or needing
   to associate with ad-tech at all.
2. We generate a single log event for a single action taken by the user. That means that if
   the user takes an action that causes an error, the error is logged _with_ that action.

   This appears to be good practice, as it allows us to associate what actions are taken
   with their result very effectively!
3. Log as little as is necessary. A messy log makes it very hard to see what's important
   so we default to not logging enough rather than logging to much. It's easy to add on
   logging, but harder to remove code.
"""
from typing import List

import analytics
# Write key taken from segement.com
analytics.write_key = '6I7ptc5wcIGC4WZ0N1t0NXvvAbjRGUgX' 

import pandas as pd
import sys
import getpass
import json

from mitosheet.user.user_utils import get_user_field
from mitosheet._version import __version__
from mitosheet.errors import get_recent_traceback_as_list
import platform # Package for getting the operating system of the user

def identify():
    """
    Helper function for identifying a user to Segment
    """
    static_user_id = get_user_field('static_user_id')
    user_email = get_user_field('user_email')

    # Get the operating system
    operating_system = platform.system()

    analytics.identify(static_user_id, {
        'location': __file__,
        'python_version': sys.version_info,
        'mito_version': __version__,
        'operating_system': operating_system,
        'email': user_email,
        'user': getpass.getuser(),
    })

def log(log_event, params=None):
    if params is None:
        params = {}

    analytics.track(
        get_user_field('static_user_id'), 
        log_event, 
        params
    )


def log_recent_error(log_event=None):
    """
    A helper function for logging the most recent error that has occured.

    log_event defaults to an execution_error_log_event
    """
    if log_event is None:
        log_event = 'execution_error_log_event'

    # We get the error, see more here: https://wiki.python.org/moin/HandlingExceptions
    e = sys.exc_info()[0]

    # If we have some other error, we just report this as an execution error
    log(
        log_event, 
        {
            'header': 'Execution Error', 
            'to_fix': 'Sorry, there was an error during executing this code.',
            'error': str(e),
            'traceback': get_recent_traceback_as_list()
        }
    )



def log_event_processed(event, widget_state_container, failed=False, edit_error=None):
    """
    Helper function for logging when an event is processed
    by the widget state container. 

    Does it's best to fill in helpful meta-data for interpreting the event
    including the state of the widget_state_container _after_ the step
    was applied.

    NOTE: if processing the event fails, then failed should be True. If there was an
    edit error that was thrown during the processing of the event, then edit_error
    should be set to that error.
    """
    try:
        # First, we get all the params of the event, and append them with _params_
        event_properties = {
            'params_' + key: value for key, value in event.items()
        }

        # If it's an import event, we also log the dataframe metadata
        # (and we also log in the case of a replayed analysis)
        if 'import' in event['type'] or 'replay_analysis_update' in event['type']:
            df_properties = get_dfs_metadata(widget_state_container.dfs)
        else:
            df_properties = {}

        # We also get some metadata about the widget state container at this state
        wsc_properties = {
            # NOTE: analysis name is the UUID that mito saves the analysis under
            'wsc_analysis_name': widget_state_container.analysis_name,
            'wsc_curr_step_idx': widget_state_container.curr_step_idx,
            'wsc_curr_step_type': widget_state_container.curr_step['step_type'],
            'wsc_df_names': widget_state_container.curr_step['df_names']
        }

        # Python properties, just for helpful data about the system
        python_properties = {
            'version_python': sys.version_info,
            'version_mito': __version__ 
        }

        # We also check there is an edit_error, and if there is, then we add the error logs
        if edit_error is not None:
            error_properties = {
                'error_type': edit_error.type_,
                'error_header': edit_error.header,
                'error_to_fix': edit_error.to_fix,
                'error_traceback': get_recent_traceback_as_list(),
            }
        elif failed:
            # Otherwise, if there is no edit_error, and we still failed, then we must have
            # gotten an execution error
            error_properties = {
                'error_type': 'execution_error',
                'error_header': 'Execution Error',
                'error_traceback': get_recent_traceback_as_list(),
            }
        else:
            error_properties = {}

        # We choose to log the event type, as it is the best high-level item for our logs
        # and we append a _failed if the event failed in doing this.
        log_event = event['type'] + ('_failed' if failed else '')

        log(
            log_event, 
            dict(
                **python_properties,
                **event_properties,
                **wsc_properties,
                **df_properties,
                **error_properties
            )
        )
    except:
        # We don't want logging to ever brick the application, so if the logging fails
        # we just log simple information about the event. This should never occur, but it
        # is just a precaution - some defensive programming so to speak
        log(
            event['type'],
            event
        )

def get_dfs_metadata(dfs: List[pd.DataFrame]):
    """
    A helper function to log metadata about a list of dataframes, 
    that does not pass any sensitive information of the dataframe
    elsewhere.
    """
    try:
        df_shapes = {f'df_{idx}_shape': {'row': df.shape[0], 'col': df.shape[1]} for idx, df in enumerate(dfs)}
        df_headers = {f'df_{idx}_headers': list(df.keys()) for idx, df in enumerate(dfs)}
        df_dtypes = {f'df_{idx}_dtypes': [str(df[key].dtype) for key in df.keys()] for idx, df in enumerate(dfs)}

        return dict(
            {'df_count': len(dfs)},
            **df_shapes,
            **df_headers,
            **df_dtypes
        )
    except:
        # We don't mind if logging fails
        pass
    return {}
